unit ftobju;

// Copyright  2002 by Ziff Davis Media, Inc.
// Written by Neil J. Rubenking
{$warn UNIT_PLATFORM off}
{$warn SYMBOL_PLATFORM off}
{$warn UNSAFE_CODE off}
{$warn UNSAFE_TYPE off}
{$warn UNSAFE_CAST off}
interface
USES Windows, SysUtils;

type
  TftObj = class(TObject) // holds file-time information
  private
    fName         : String;
    fStatus       : String;
    fAttrOrg      : DWORD;    //original attributes, never changed
    fAttrWork     : DWORD;    //working attributes
    fAttrPending  : DWORD;    //new attributes
    ftCreOrg      : TFileTime;//original creation time, never changed
    ftCreWork     : TFileTime;//working creation time
    ftCrePending  : TFileTime;//new creation time, initially blank
    ftModOrg      : TFileTime;//original modified time, never changed
    ftModWork     : TFileTime;//working modified time
    ftModPending  : TFileTime;//new modified time, initially blank
    fMatchStatus  : Integer;
    procedure SetDCrePending(const Value: TDateTime);
    procedure SetDModPending(const Value: TDateTime);
    function GetSCreTime: String;
    function GetSModTime: String;
    function GetSAttr: String;
    function GetSAttrPending: String;
    function AttString(D : DWORD) : String;
    procedure ClearModPending;
    procedure ClearCrePending;
    function DateString(const Value : TDateTime) : String;
  public
    class function FileTime2TDateTime(T: TFileTime): TDateTime;
    class function DateTime2FileTime(D : TDateTime): TFileTime;
    constructor Create(vAttr : DWORD; const vName : String; vCreOrg,
      vModOrg : TFileTime);
    procedure ClearPending;
    function PerformPending : Integer;
    function Undo           : Integer;
    procedure SetModPendingOpt(D : TDateTime; T : TDateTime);
    procedure SetCrePendingOpt(D : TDateTime; T : TDateTime);
    procedure SetFromFolder(const S : String; doMod,
      doCre : Boolean);
    function AttrChange(vRem, vAdd : DWORD) : String;
    function Report : String;
    property Filename     : String    read fName;
    property Status       : String    read fStatus;
    property SCreTime     : String    read GetSCreTime;
    property SModTime     : String    read GetSModTime;
    property Attr         : DWORD     read fAttrWork;
    property SAttr        : String    read GetSAttr;
    property MatchStatus  : Integer   read fMatchStatus
                                      write fMatchStatus;
    property CreTime      : TFileTime read ftCreWork
                                      write ftCreWork;
    property CrePending   : TFileTime read ftCrePending
                                      write ftCrePending;
    property ModTime      : TFileTime read ftModWork
                                      write ftModWork;
    property ModPending   : TFileTime read ftModPending
                                      write ftModPending;
  end;

const
  err_file = $01;
  err_mod  = $02;
  err_cre  = $04;
  err_att  = $08;

implementation
uses TFuncs2;

procedure TftObj.SetDCrePending(const Value: TDateTime);
// Set the pending Created time from a TDateTime
begin
  ftCrePending := DateTime2FileTime(Value);
end;

procedure TftObj.SetDModPending(const Value: TDateTime);
// Set the pending Modified time from a TDateTime
begin
  ftModPending := DateTime2FileTime(Value);
end;

function TftObj.GetSCreTime: String;
// Get the current Created time as a string
begin
  Result := DateString(FileTime2TDateTime(ftCreWork));
end;

function TftObj.GetSModTime: String;
// Get the current Modified time as a string
begin
  Result := DateString(FileTime2TDateTime(ftModWork));
end;

function TftObj.GetSAttr: String;
// Get the current attributes as a string
begin
  Result := AttString(fAttrWork);
end;

function TftObj.GetSAttrPending: String;
// Get the pending attributes as a string
begin
  Result := AttString(fAttrPending);
end;

function TftObj.AttString(D: DWORD): String;
// Return a string representing the specified attributes
begin
  IF D = DWORD(-1) THEN Result := ''
  ELSE
    begin
      Result := '     ';
      IF D AND FILE_ATTRIBUTE_ARCHIVE   > 0 THEN Result[1] := 'A';
      IF D AND FILE_ATTRIBUTE_SYSTEM    > 0 THEN Result[2] := 'S';
      IF D AND FILE_ATTRIBUTE_HIDDEN    > 0 THEN Result[3] := 'H';
      IF D AND FILE_ATTRIBUTE_READONLY  > 0 THEN Result[4] := 'R';
      IF D AND FILE_ATTRIBUTE_DIRECTORY > 0 THEN Result[5] := 'D';
    end;
end;

procedure TftObj.ClearModPending;
// A value of all $FF means nothing pending
begin
  FillChar(ftModPending, SizeOf(ftModPending), $FF);
end;

procedure TftObj.ClearCrePending;
// A value of all $FF means nothing pending
begin
  FillChar(ftCrePending, SizeOf(ftCrePending), $FF);
end;

function TftObj.DateString(const Value: TDateTime): String;
begin
  IF Value < 0 THEN
    Result := '<empty>'
  ELSE
    Result := FormatDateTime(wholeFmt, Value);
end;

class function TftObj.FileTime2TDateTime(T: TFileTime): TDateTime;
VAR
  FLoc : TFileTime;
  STmp : TSystemTime;
begin
  Result := -1;
  IF NOT FileTimeToLocalFileTime(T, FLoc) THEN Exit;
  IF NOT FileTimeToSystemTime(FLoc, STmp) THEN Exit;
  try
    Result := SystemTimeToDateTime(STmp);
  except
    ON EConvertError DO Result := -1;
  end;
end;

class function TftObj.DateTime2FileTime(D : TDateTime): TFileTime;
VAR
  SysTim  : TSystemTime;
  LocTim  : TFileTime;
  FileTim : TFileTime;
begin
  FillChar(Result, SizeOf(Result), $FF);
  DateTimeToSystemTime(D, SysTim);
  IF SystemTimeToFileTime(SysTim, LocTim) THEN
    IF LocalFileTimeToFileTime(LocTim, FileTim) THEN
      Result := FileTim;
end;

constructor TftObj.Create(vAttr : DWORD; const vName : String;
  vCreOrg, vModOrg: TFileTime);
begin
  Inherited Create;
  fAttrOrg     := vAttr;
  fAttrWork    := vAttr;
  fName        := vName;
  fStatus      := '';
  ftCreOrg     := vCreOrg;
  ftModOrg     := vModOrg;
  ftCreWork    := vCreOrg;
  ftModWork    := vModOrg;
  ClearPending;
end;

procedure TftObj.ClearPending;
begin
  ClearCrePending;
  ClearModPending;
  fAttrPending := DWORD(-1);
end;

function TftObj.PerformPending : Integer;
VAR
  PMod   : PFileTime; // Possibily-nil pointer to new Mod time
  PCre   : PFileTime; // Possibily-nil pointer to new Cre time
  NuMod  : TFileTime; // For sanity-check after change
  NuCre  : TFileTime; // For sanity-check after change
  fHan   : THandle;   // Handle of file/folder
  Poss   : Integer;   // Used in error handling
  justRO : Boolean;
begin
  fStatus := '';
  Result  := 0;
  Poss    := 0;
  justRO  := False;
  IF (Int64(ftModPending) <> -1) OR (Int64(ftCrePending) <> -1) THEN
    begin
      // If the file is read-only, we need to temporarily remove the
      // read-only attribute so we can change the times
      IF fAttrWork AND FILE_ATTRIBUTE_READONLY > 0 THEN
        begin
          IF fAttrPending = DWORD(-1) THEN
            begin
              fAttrPending := fAttrWork;
              justRO := True;
            end;
          SetFileAttributes(PChar(fName),
            fAttrWork AND (NOT FILE_ATTRIBUTE_READONLY));
        end;
      IF Int64(ftModPending) = -1 THEN
        PMod := nil
      ELSE
        begin
          PMod := @ftModPending;
          Poss := Poss OR err_mod;
        end;
      IF Int64(ftCrePending) = -1 THEN
        PCre := nil
      ELSE
        begin
          PCre := @ftCrePending;
          Poss := Poss OR err_cre;
        end;
      IF (Win32Platform = VER_PLATFORM_WIN32_NT) AND
        (fAttrWork AND FILE_ATTRIBUTE_DIRECTORY > 0) THEN
          // These flags allow the program to manipulate folder
          // date/time stamps, but only in NT-family
        fHan := CreateFile(PChar(fName), GENERIC_READ OR
          GENERIC_WRITE, FILE_SHARE_READ OR FILE_SHARE_WRITE,
          nil, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, 0)
      ELSE
        fHan := CreateFile(PChar(fName), GENERIC_READ OR
          GENERIC_WRITE, FILE_SHARE_READ OR FILE_SHARE_WRITE,
          nil, OPEN_EXISTING, 0, 0);
      // Can't change date/times for folder if not NT-family
      IF (Win32Platform <> VER_PLATFORM_WIN32_NT) AND
        (fAttrWork AND FILE_ATTRIBUTE_DIRECTORY > 0) THEN
        begin
          Result := Result OR Poss;
          fStatus := fStatus + #13#10'Folder date/time cannot be '+
            'changed in Windows 9x';
        end
      // Couldn't open file? Then can't change time(s)
      ELSE IF fHan = HFILE_ERROR THEN
        begin
          Result := Result OR Poss;
          fStatus := fStatus + Format(#13#10'Unable to open file: '+
            '"%s"', [SysErrorMessage(GetLastError)]);
        end
      ELSE
        begin
          try
            // Tryp to set the time(s)
            IF NOT SetFileTime(fHan, PCre, nil, PMod) THEN
              begin
                Result := Result OR Poss;
                fStatus := fStatus + Format(#13#10'Unable to set '+
                  'file time: "%s"',[SysErrorMessage(GetLastError)]);
              end
            ELSE
              begin
                // Now get the actual time(s)
                GetFileTime(fHan, @nuCre, nil, @nuMod);
                IF Int64(ftModPending) <> -1 THEN
                  begin
                    IF CompFileTime(ftModPending, ftModOrg) = 0 THEN
                      fStatus := fStatus + Format(#13#10'Modified '+
                        'date/time same as original of %s',
                        [DateString(FileTime2TDateTime(ftModOrg))])
                    ELSE
                      fStatus := fStatus + Format(#13#10'Modified '+
                        'date/time changed from original of %s',
                        [DateString(FileTime2TDateTime(ftModOrg))]);
                    ftModWork := nuMod;
                    ClearModPending;
                  end;
                IF Int64(ftCrePending) <> -1 THEN
                  begin
                    IF CompFileTime(ftCrePending, ftCreOrg) = 0 THEN
                      fStatus := fStatus + Format(#13#10'Created '+
                        'date/time same as original of %s',
                        [DateString(FileTime2TDateTime(ftCreOrg))])
                    ELSE
                      fStatus := fStatus + Format(#13#10'Created '+
                        'date/time changed from original of %s',
                        [DateString(FileTime2TDateTime(ftCreOrg))]);
                    ftCreWork := nuCre;
                    ClearCrePending;
                  end;
              end;
          finally
            _lClose(fHan);
          end;
        end;
    end;
  IF fAttrPending <> DWORD(-1) THEN
    begin
      IF SetFileAttributes(PChar(fName), fAttrPending) THEN
        begin
          // "justRO" means we were not asked to change attributes,
          // but had to temporarily remove read-only.
          IF NOT justRO THEN
            begin
              IF fAttrOrg = fAttrPending THEN
                fStatus := fStatus + Format(#13#10'Attributes same '+
                  'as original of "%s"', [AttString(fAttrOrg)])
              ELSE
                fStatus := fStatus + Format(#13#10'Attributes '+
                  'changed from original of "%s"',
                  [AttString(fAttrOrg)]);
            end;
          fAttrWork := GetFileAttributes(PChar(fName));
          fAttrPending := DWORD(-1);
        end
      ELSE
        begin
          Result := Result OR err_att;
          fStatus := fStatus + Format(#13#10'Unable to set '+
            'attributes: "%s"', [SysErrorMessage(GetLastError)]);
        end;
    end;
  IF fStatus <> '' THEN System.Delete(fStatus, 1, 2);
end;

function TftObj.Undo : Integer;
VAR
  PMod     : PFileTime; // pointer to original Modified time
  PCre     : PFileTime; // pointer to original Created time
  ftModNOW : TFileTime; // actual current Modified time
  ftCreNOW : TFileTime; // actual current Created time
  fAttrNOW : DWORD;     // actual current Attributes
  fHan     : THandle;   // handle for file/folder
  Poss     : Integer;   // used in error handling
  WasRO    : Boolean;   // because we must remove R/O to change time
begin
  Result  := 0;
  Poss    := 0;
  fStatus := '';
  WasRO   := False;
  fAttrNOW := GetFileAttributes(PChar(fName));
  IF Win32Platform = VER_PLATFORM_WIN32_NT THEN
    fHan := CreateFile(PChar(fName), GENERIC_READ OR GENERIC_WRITE,
      FILE_SHARE_READ OR FILE_SHARE_WRITE, nil, OPEN_EXISTING,
      FILE_FLAG_BACKUP_SEMANTICS, 0)
  ELSE
    fHan := CreateFile(PChar(fName), GENERIC_READ OR GENERIC_WRITE,
      FILE_SHARE_READ OR FILE_SHARE_WRITE, nil, OPEN_EXISTING,
      0, 0);
  IF fHan = HFILE_ERROR THEN
    begin
      ftModNOW := ftModWork;
      ftCreNOW := ftCreWork;
    end
  ELSE
    try
      GetFileTime(fHan, @ftCreNOW, nil, @ftModNOW);
    finally
      _lClose(fHan);
    end;
  IF CompFileTime(ftModNOW, ftModOrg) = 0 THEN
    PMod := nil
  ELSE
    begin
      PMod := @ftModOrg;
      Poss := Poss OR err_mod;
    end;
  IF CompFileTime(ftCreNOW, ftCreOrg) = 0 THEN
    PCre := nil
  ELSE
    begin
      PCre := @ftCreOrg;
      Poss := Poss OR err_cre;
    end;
  IF (PMod <> nil) OR (PCre <> nil) THEN
    begin
      // Remove read-only if necessary
      IF fAttrNOW AND FILE_ATTRIBUTE_READONLY > 0 THEN
        begin
          WasRO := True;
          SetFileAttributes(PChar(fName),
            fAttrNOW AND (NOT FILE_ATTRIBUTE_READONLY));
        end;
      IF Win32Platform = VER_PLATFORM_WIN32_NT THEN
        fHan := CreateFile(PChar(fName), GENERIC_READ OR
          GENERIC_WRITE, FILE_SHARE_READ OR FILE_SHARE_WRITE,
          nil, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, 0)
      ELSE
        fHan := CreateFile(PChar(fName), GENERIC_READ OR
          GENERIC_WRITE, FILE_SHARE_READ OR FILE_SHARE_WRITE, nil,
          OPEN_EXISTING, 0, 0);
      IF fHan = HFILE_ERROR THEN
        begin
          Result := Result OR Poss;
          fStatus := fStatus + Format(#13#10'Unable to open file: '+
            '"%s"', [SysErrorMessage(GetLastError)]);
        end
      ELSE
        begin
          try
            IF NOT SetFileTime(fHan, PCre, nil, PMod) THEN
              begin
                Result := Result OR Poss;
                fStatus := fStatus + Format(#13#10'Unable to set '+
                  'file time: "%s"',[SysErrorMessage(GetLastError)]);
              end
            ELSE
              begin
                GetFileTime(fHan, @ftCreNow, nil, @ftModNow);
                ftModWork := ftModNow;
                ftCreWork := ftCreNow;
                ClearModPending;
                ClearCrePending;
              end;
          finally
            _lClose(fHan);
          end;
        end;
    end;
  IF WasRO OR (fAttrNOW <> fAttrOrg) THEN
    begin
      IF SetFileAttributes(PChar(fName), fAttrOrg) THEN
        begin
          fAttrWork := GetFileAttributes(PChar(fName));
          fAttrPending := DWORD(-1);
        end
      ELSE
        begin
          Result := Result OR err_att;
          fStatus := fStatus + Format(#13#10'Unable to set '+
            'attributes: "%s"', [SysErrorMessage(GetLastError)]);
        end;
    end;
  IF fStatus <> '' THEN System.Delete(fStatus, 1, 2);
end;

procedure TftObj.SetModPendingOpt(D, T: TDateTime);
// Set the pending Modified date/time. If either D or T is -1, it
// means the corresponding control in the main window was NOT
// checked - so use the current value for that one.
begin
  IF D = -1 THEN D := FileTime2TDateTime(ftModWork);
  IF T = -1 THEN T := FileTime2TDateTime(ftModWork);
  SetDModPending(Int(D)+Frac(T));
end;

procedure TftObj.SetCrePendingOpt(D, T: TDateTime);
// Set the pending Created date/time. If either D or T is -1, it
// means the corresponding control in the main window was NOT
// checked - so use the current value for that one.
begin
  IF D = -1 THEN D := FileTime2TDateTime(ftCreWork);
  IF T = -1 THEN T := FileTime2TDateTime(ftCreWork);
  SetDCrePending(Int(D)+Frac(T));
end;

procedure TftObj.SetFromFolder(const S: String; doMod,
  doCre: Boolean);
VAR
  FindData : TWin32FindData;
  FindHan  : THandle;
  Path     : String;
begin
  IF NOT (doMod OR doCre) THEN Exit;
  CASE fMatchStatus OF
    2 : Path := IncludeTrailingBackslash(S) + ExtractFilename(fName);
    1 : Path := IncludeTrailingBackslash(S) +
      ChangeFileExt(ExtractFilename(fName), '.*');
    ELSE Exit;
  END;
  FindHan := FindFirstFile(PChar(Path), FindData);
  IF FindHan = INVALID_HANDLE_VALUE THEN Exit
  ELSE
    try
      WITH FindData DO
        begin
          IF doMod THEN ModPending := ftLastWriteTime;
          IF doCre THEN CrePending := ftCreationTime;
        end;
    finally
      Windows.FindClose(FindHan);
    end;
end;

function TftObj.AttrChange(vRem, vAdd: DWORD): String;
begin
  fAttrPending := (fAttrWork AND vRem) OR vAdd;
  Result := GetSAttrPending;
end;

function TftObj.Report: String;
begin
  Result := Format(#9'M: %s'#9'C: %s'#9'%s',
    [SModTime, SCreTime, SAttr]);
end;

end.
